from contextlib import redirect_stdout
from pathlib import Path
from traceback import extract_stack
from typing import NamedTuple, Iterable, Optional
from zipfile import ZIP_DEFLATED, ZipFile

from git import InvalidGitRepositoryError, Repo, IndexObject
from IPython import InteractiveShell, get_ipython


class ZipPaths(NamedTuple):
    script_file_path: Path
    notebook_file_path: Optional[Path]
    zip_file_path: Path
    script_file_path_in_zip: Path
    notebook_file_path_in_zip: Optional[Path]


def _create_zip_archive(
    metadata_directory_path: Path,
    base_directory_path: Path,
    script_file_path: Path,
    notebook_file_path: Optional[Path],
    other_paths: Iterable[Path] = tuple(),
) -> ZipPaths:
    all_paths = {
        script_file_path,
        *([notebook_file_path] if notebook_file_path else []),
        *other_paths,
    }
    zip_file_path = metadata_directory_path / "code.zip"
    with ZipFile(zip_file_path, "w", compression=ZIP_DEFLATED) as zip_file:
        for path in all_paths:
            relative_path = path.relative_to(base_directory_path)
            zip_file.write(path, arcname=relative_path)
    return ZipPaths(
        script_file_path=script_file_path,
        notebook_file_path=notebook_file_path,
        zip_file_path=zip_file_path,
        script_file_path_in_zip=script_file_path.relative_to(base_directory_path),
        notebook_file_path_in_zip=(
            notebook_file_path.relative_to(base_directory_path)
            if notebook_file_path
            else None
        ),
    )


def _create_notebook_zip_archive(
    metadata_directory_path: Path,
    ipython: InteractiveShell,
) -> ZipPaths:
    notebook_file_path = metadata_directory_path / "notebook.ipynb"
    script_file_path = metadata_directory_path / "script.py"
    with redirect_stdout(None):
        ipython.run_line_magic("save", f"-f {script_file_path} 1-9999")
        ipython.run_line_magic("notebook", str(notebook_file_path))
    return _create_zip_archive(
        metadata_directory_path=metadata_directory_path,
        base_directory_path=metadata_directory_path,
        script_file_path=script_file_path,
        notebook_file_path=notebook_file_path,
    )


def create_git_zip_archive(
    metadata_directory_path: Path,
    script_file_path: Path,
) -> Optional[ZipPaths]:
    """
    Creates a ZIP archive containing all tracked files in a given Git repository.

    :param directory_path_in_git_repository: A directory that is inside the Git repository that should be archived.
    """

    try:
        repo = Repo(script_file_path, search_parent_directories=True)
    except InvalidGitRepositoryError:
        return None

    working_tree_dir = repo.working_tree_dir
    if working_tree_dir is None:
        return None
    working_tree_dir_path = Path(working_tree_dir)

    tracked_paths = (
        Path(item.abspath)
        for item in repo.commit().tree.traverse()
        if isinstance(item, IndexObject)
    )
    tracked_file_paths = (
        path
        for path in tracked_paths
        if path.is_file()
    )

    return _create_zip_archive(
        metadata_directory_path=metadata_directory_path,
        base_directory_path=working_tree_dir_path,
        script_file_path=script_file_path,
        notebook_file_path=None,
        other_paths=tracked_file_paths,
    )


def create_code_archive(
    metadata_directory_path: Path,
) -> ZipPaths:
    ipython = get_ipython()
    if ipython is not None:
        return _create_notebook_zip_archive(
            metadata_directory_path=metadata_directory_path,
            ipython=ipython,
        )
    else:
        script_file_path = Path(extract_stack()[0].filename).resolve()

        paths = create_git_zip_archive(
            metadata_directory_path=metadata_directory_path,
            script_file_path=script_file_path,
        )
        if paths is not None:
            return paths
        return _create_zip_archive(
            metadata_directory_path=metadata_directory_path,
            base_directory_path=script_file_path.parent,
            script_file_path=script_file_path,
            notebook_file_path=None,
        )



from io import BytesIO
from os import environ
from pathlib import Path
from shutil import copy as shutil_copy
from subprocess import check_output  # nosec
from sys import executable
from tempfile import TemporaryDirectory
from zipfile import ZipFile

from ruamel.yaml import YAML

_EXAMPLES_DIR_PATH = Path(__file__).parent.parent / "examples"
_EXAMPLE_SCRIPT_PATH = _EXAMPLES_DIR_PATH / "example.py"
_EXAMPLE_NOTEBOOK_PATH = _EXAMPLES_DIR_PATH / "example-notebook.ipynb"


def test_python_script() -> None:
    with TemporaryDirectory() as temp_dir:
        working_dir_path = Path(temp_dir)

        # Copy the example script to the working directory.
        shutil_copy(_EXAMPLE_SCRIPT_PATH, working_dir_path / "example.py")

        # Set the PYTHONPATH environment variable to include the parent directory of the `tirex_tracker` package.
        env = environ.copy()
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] += ":" + str(Path(__file__).parent.parent.resolve())

        export_file_path = working_dir_path / "ir_metadata.yml"
        assert not export_file_path.exists()

        # Run the script using the `python3` command.
        check_output(
            args=[executable, "example.py"],
            cwd=working_dir_path,
            env=env,
        )  # nosec

        assert export_file_path.exists()

        buffer = Path(export_file_path).read_bytes()
        if buffer.startswith(b"ir_metadata.start\n"):
            buffer = buffer[len(b"ir_metadata.start\n") :]
        if buffer.endswith(b"ir_metadata.end\n"):
            buffer = buffer[: -len(b"ir_metadata.end\n")]

        with BytesIO(buffer) as yaml_file:
            yaml = YAML(typ="safe")
            yaml_content = yaml.load(yaml_file)

        assert "implementation" in yaml_content
        assert "script" in yaml_content["implementation"]
        assert "path" in yaml_content["implementation"]["script"]
        assert "example.py" in yaml_content["implementation"]["script"]["path"]
        assert "source" in yaml_content["implementation"]
        assert "archive" in yaml_content["implementation"]["source"]
        assert "path" in yaml_content["implementation"]["source"]["archive"]
        assert "script path" in yaml_content["implementation"]["source"]["archive"]
        assert "example.py" in yaml_content["implementation"]["source"]["archive"]["script path"]

        archive_path = Path(yaml_content["implementation"]["source"]["archive"]["path"])
        assert archive_path.exists()

        with ZipFile(archive_path) as zip_file:
            archive_names = zip_file.namelist()
        assert "example.py" in archive_names
        assert len(archive_names) == 1


def test_jupyter_notebook() -> None:
    with TemporaryDirectory() as temp_dir:
        working_dir_path = Path(temp_dir)

        # Copy the example notebook to the working directory.
        shutil_copy(_EXAMPLE_NOTEBOOK_PATH, working_dir_path / "example-notebook.ipynb")

        # Set the PYTHONPATH environment variable to include the parent directory of the `tirex_tracker` package.
        env = environ.copy()
        if "PYTHONPATH" in env:
            env["PYTHONPATH"] += ":" + str(Path(__file__).parent.parent.resolve())

        export_file_path = working_dir_path / "ir_metadata.yml"
        assert not export_file_path.exists()

        # Run the notebook using the `runnb` command.
        check_output(
            args=["runnb", "--allow-not-trusted", "example-notebook.ipynb"],
            cwd=working_dir_path,
            env=env,
        )  # nosec

        assert export_file_path.exists()

        buffer = Path(export_file_path).read_bytes()
        if buffer.startswith(b"ir_metadata.start\n"):
            buffer = buffer[len(b"ir_metadata.start\n") :]
        if buffer.endswith(b"ir_metadata.end\n"):
            buffer = buffer[: -len(b"ir_metadata.end\n")]

        with BytesIO(buffer) as yaml_file:
            yaml = YAML(typ="safe")
            yaml_content = yaml.load(yaml_file)

        assert "implementation" in yaml_content
        assert "script" in yaml_content["implementation"]
        assert "path" in yaml_content["implementation"]["script"]
        assert "script.py" in yaml_content["implementation"]["script"]["path"]
        assert "notebook" in yaml_content["implementation"]
        assert "path" in yaml_content["implementation"]["notebook"]
        assert "notebook.ipynb" in yaml_content["implementation"]["notebook"]["path"]
        assert "source" in yaml_content["implementation"]
        assert "archive" in yaml_content["implementation"]["source"]
        assert "path" in yaml_content["implementation"]["source"]["archive"]
        assert "script path" in yaml_content["implementation"]["source"]["archive"]
        assert "script.py" in yaml_content["implementation"]["source"]["archive"]["script path"]
        assert "notebook path" in yaml_content["implementation"]["source"]["archive"]
        assert "notebook.ipynb" in yaml_content["implementation"]["source"]["archive"]["notebook path"]

        archive_path = Path(yaml_content["implementation"]["source"]["archive"]["path"])
        assert archive_path.exists()

        with ZipFile(archive_path) as zip_file:
            archive_names = zip_file.namelist()
        assert "script.py" in archive_names
        assert "notebook.ipynb" in archive_names
        assert len(archive_names) == 2



from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import ZipFile

from tirex_tracker.archive_utils import create_git_zip_archive

_EXAMPLE_GIT_REPOSITORY_PATH = (
    Path(__file__).parent / "resources" / "example-git-repositories.zip"
)


def test_code_extraction_from_git_repository() -> None:
    expected_code_files = {
        "some-directory/.gitignore",
        "some-directory/Dockerfile",
        "some-directory/script.sh",
    }

    with TemporaryDirectory() as temp_working_dir:
        working_directory_path = Path(temp_working_dir)

        # Extract the example Git repository to the working directory.
        with ZipFile(_EXAMPLE_GIT_REPOSITORY_PATH, "r") as zip_file:
            zip_file.extractall(working_directory_path)

        script_file_path = (
            working_directory_path / "git-repo-clean" / "some-directory" / "script.sh"
        )
        script_file_path.chmod(0o0766)  # nosec

        with TemporaryDirectory() as tmp_metadata_dir:
            metadata_dir_path = Path(tmp_metadata_dir)

            actual = create_git_zip_archive(
                metadata_directory_path=metadata_dir_path,
                script_file_path=script_file_path,
            )

            assert actual is not None
            assert actual.zip_file_path.exists()

            with ZipFile(actual.zip_file_path) as zip_file:
                archive_names = zip_file.namelist()

            assert sorted(archive_names) == sorted(expected_code_files)
